"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeterministicFiniteAutomaton = void 0;
const FiniteAutomaton_1 = require("./FiniteAutomaton");
class DeterministicFiniteAutomaton extends FiniteAutomaton_1.FiniteAutomaton {
    constructor(testLogic, automaton, automatonId) {
        super(testLogic, automaton, 'deterministic', automatonId);
    }
    #generateMergedDfaData(state, newStates, newTransitions, newFinalStates, mergeOperation, inputAutomaton, isComposite, separator) {
        const currentAutomatonFinalStates = new Set(this.automaton.final_states);
        const inputAutomatonFinalStates = new Set(inputAutomaton ? inputAutomaton.automaton.final_states : []);
        this.automaton.states.forEach((currentDfaState) => {
            const newState = isComposite ? `${currentDfaState}${separator}${state}` : currentDfaState;
            if (isComposite) {
                newStates.push(newState);
                newTransitions[newState] = {};
                this.automaton.alphabets.forEach((automatonAlphabet) => {
                    newTransitions[newState][automatonAlphabet] = [
                        this.automaton.transitions[currentDfaState][automatonAlphabet] +
                            separator +
                            (inputAutomaton
                                ? inputAutomaton.automaton.transitions[state][automatonAlphabet]
                                : ''),
                    ];
                });
                if (mergeOperation === 'or' &&
                    ((inputAutomaton ? inputAutomatonFinalStates.has(state) : true) ||
                        currentAutomatonFinalStates.has(currentDfaState))) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'and' &&
                    (inputAutomaton ? inputAutomatonFinalStates.has(state) : true) &&
                    currentAutomatonFinalStates.has(currentDfaState)) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'not' && !currentAutomatonFinalStates.has(currentDfaState)) {
                    newFinalStates.add(newState);
                }
            }
            else {
                if (mergeOperation === 'or' &&
                    ((inputAutomaton ? inputAutomatonFinalStates.has(newState) : true) ||
                        currentAutomatonFinalStates.has(newState))) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'and' &&
                    (inputAutomaton ? inputAutomatonFinalStates.has(newState) : true) &&
                    currentAutomatonFinalStates.has(newState)) {
                    newFinalStates.add(newState);
                }
                else if (mergeOperation === 'not' && !currentAutomatonFinalStates.has(newState)) {
                    newFinalStates.add(newState);
                }
            }
        });
    }
    #merge(finiteAutomaton, mergeOperation, mergedDfaOptions) {
        const { separator = '.', label, description } = mergedDfaOptions ?? {};
        const newStates = [];
        const newTransitions = {};
        const isComposite = Boolean(finiteAutomaton &&
            (finiteAutomaton ? finiteAutomaton.getAutomatonId() : '') !== this.getAutomatonId());
        const newDfaId = isComposite && finiteAutomaton
            ? this.getAutomatonId() + separator + finiteAutomaton.getAutomatonId()
            : this.getAutomatonId();
        const newStartState = isComposite && finiteAutomaton
            ? this.automaton.start_state + separator + finiteAutomaton.automaton.start_state
            : this.automaton.start_state;
        const newFinalStates = new Set();
        if (finiteAutomaton) {
            finiteAutomaton.automaton.states.forEach((state) => {
                this.#generateMergedDfaData(state, newStates, newTransitions, newFinalStates, mergeOperation, finiteAutomaton, isComposite, separator);
            });
        }
        else {
            this.#generateMergedDfaData('', newStates, newTransitions, newFinalStates, mergeOperation, finiteAutomaton, isComposite, separator);
        }
        return new DeterministicFiniteAutomaton((inputString, automatonTestResult) => {
            if (mergeOperation === 'or') {
                return (finiteAutomaton.testLogic(inputString, automatonTestResult) ||
                    this.testLogic(inputString, automatonTestResult));
            }
            else if (mergeOperation === 'and') {
                return (finiteAutomaton.testLogic(inputString, automatonTestResult) &&
                    this.testLogic(inputString, automatonTestResult));
            }
            else {
                return !this.testLogic(inputString, automatonTestResult);
            }
        }, {
            final_states: Array.from(newFinalStates),
            label: label ??
                mergeOperation +
                    '(' +
                    `${this.automaton.label}${mergeOperation !== 'not' ? ', ' : ''}${finiteAutomaton ? finiteAutomaton.automaton.label : ''}` +
                    ')',
            description: description ??
                mergeOperation.toUpperCase() +
                    '(' +
                    `${this.automaton.description}${mergeOperation !== 'not' ? ', ' : ''}${finiteAutomaton ? finiteAutomaton.automaton.description : ''}` +
                    ')',
            start_state: isComposite ? newStartState : this.automaton.start_state,
            states: isComposite ? newStates : JSON.parse(JSON.stringify(this.automaton.states)),
            transitions: (isComposite
                ? newTransitions
                : JSON.parse(JSON.stringify(this.automaton.transitions))),
            alphabets: this.automaton.alphabets,
            epsilon_transitions: null,
        }, isComposite ? newDfaId : this.getAutomatonId());
    }
    AND(dfaModule, mergedDfaOptions) {
        return this.#merge(dfaModule, 'and', mergedDfaOptions);
    }
    NOT(mergedDfaOptions) {
        return this.#merge(undefined, 'not', mergedDfaOptions);
    }
    OR(dfaModule, mergedDfaOptions) {
        return this.#merge(dfaModule, 'or', mergedDfaOptions);
    }
    generateStateGroupsRecord(stateGroups) {
        const stateGroupsRecord = {};
        const allStatesSet = new Set(this.automaton.states);
        stateGroups.forEach((stateGroup, stateGroupIndex) => {
            stateGroup.forEach((state) => {
                if (allStatesSet.has(state)) {
                    stateGroupsRecord[state] = stateGroupIndex;
                }
            });
        });
        return stateGroupsRecord;
    }
    generateEquivalenceStates(stateGroups) {
        const stateGroupsRecord = this.generateStateGroupsRecord(stateGroups);
        const stateGroupsSymbolsRecord = {};
        const singleStateGroups = [];
        const compositeStateGroups = [];
        stateGroups.forEach((stateGroup) => {
            if (stateGroup.length > 1) {
                compositeStateGroups.push(stateGroup);
            }
            else {
                singleStateGroups.push(stateGroup);
            }
        });
        compositeStateGroups.forEach((compositeStateGroup) => {
            compositeStateGroup.forEach((state) => {
                let stateGroup = '';
                this.automaton.alphabets.forEach((symbol) => {
                    stateGroup += stateGroupsRecord[this.automaton.transitions[state][symbol].toString()];
                });
                if (!stateGroupsSymbolsRecord[stateGroup]) {
                    stateGroupsSymbolsRecord[stateGroup] = [state];
                }
                else {
                    stateGroupsSymbolsRecord[stateGroup].push(state);
                }
            });
        });
        return Object.values(stateGroupsSymbolsRecord).concat(singleStateGroups);
    }
    checkEquivalenceBetweenStatesGroups(statesGroups) {
        const [statesGroupsOne, statesGroupsTwo] = statesGroups;
        if (statesGroupsOne.length !== statesGroupsTwo.length)
            return false;
        let isEquivalent = true;
        const stateGroupsSet = new Set();
        statesGroupsOne.forEach((statesGroup) => {
            stateGroupsSet.add(statesGroup.join(''));
        });
        for (let index = 0; index < statesGroupsTwo.length; index++) {
            const statesGroup = statesGroupsTwo[index];
            if (!stateGroupsSet.has(statesGroup.join(''))) {
                isEquivalent = false;
                break;
            }
        }
        return isEquivalent;
    }
    minimize(minimizedDfaOptions) {
        const finalStatesSet = new Set(this.automaton.final_states);
        const nonFinalStates = [];
        this.automaton.states.forEach((state) => {
            if (!finalStatesSet.has(state)) {
                nonFinalStates.push(state);
            }
        });
        const finalStateString = this.automaton.final_states.join('');
        let currentEquivalentStatesGroups = [nonFinalStates, this.automaton.final_states];
        let previousEquivalentStatesGroups = [];
        let shouldStop = false;
        while (!shouldStop) {
            previousEquivalentStatesGroups = currentEquivalentStatesGroups;
            currentEquivalentStatesGroups = this.generateEquivalenceStates(currentEquivalentStatesGroups);
            shouldStop = this.checkEquivalenceBetweenStatesGroups([
                currentEquivalentStatesGroups,
                previousEquivalentStatesGroups,
            ]);
        }
        const stateGroupsRecord = this.generateStateGroupsRecord(currentEquivalentStatesGroups);
        const newStateState = currentEquivalentStatesGroups.find((stateGroup) => stateGroup.includes(this.automaton.start_state));
        const newTransitions = {};
        currentEquivalentStatesGroups.forEach((currentEquivalentStatesGroup) => {
            const [firstState] = currentEquivalentStatesGroup;
            const newState = currentEquivalentStatesGroup.join('');
            newTransitions[newState] = {};
            this.automaton.alphabets.forEach((symbol) => {
                newTransitions[newState][symbol] = [
                    currentEquivalentStatesGroups[stateGroupsRecord[this.automaton.transitions[firstState][symbol][0]]].join(''),
                ];
            });
        });
        return new DeterministicFiniteAutomaton(this.testLogic, {
            label: minimizedDfaOptions?.label ?? this.automaton.label,
            alphabets: this.automaton.alphabets,
            description: minimizedDfaOptions?.description ?? this.automaton.description,
            final_states: [finalStateString],
            start_state: newStateState?.join('') ?? this.automaton.start_state,
            states: currentEquivalentStatesGroups.map((currentEquivalentStatesGroup) => currentEquivalentStatesGroup.join('')),
            transitions: newTransitions,
            epsilon_transitions: null,
        });
    }
}
exports.DeterministicFiniteAutomaton = DeterministicFiniteAutomaton;
