"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FiniteAutomaton = void 0;
const colors_1 = __importDefault(require("colors"));
const shortid_1 = __importDefault(require("shortid"));
class FiniteAutomaton {
    testLogic;
    automaton;
    #automatonId;
    #automatonType;
    constructor(testLogic, finiteAutomaton, automatonType, automatonId) {
        this.#automatonType = automatonType;
        this.#automatonId = automatonId ?? (0, shortid_1.default)();
        this.testLogic = testLogic;
        this.#normalize(finiteAutomaton);
    }
    #normalize(finiteAutomaton) {
        this.#validate(finiteAutomaton.label ?? '', this.#generatePreNormalizationErrors(finiteAutomaton));
        const appendedString = finiteAutomaton.append ?? '';
        finiteAutomaton.final_states = finiteAutomaton.final_states.map((finalState) => appendedString + finalState.toString());
        finiteAutomaton.alphabets = finiteAutomaton.alphabets.map((alphabet) => alphabet.toString());
        finiteAutomaton.start_state = appendedString + finiteAutomaton.start_state.toString();
        finiteAutomaton.states = finiteAutomaton.states.map((state) => appendedString + state.toString());
        if (finiteAutomaton.epsilon_transitions) {
            Object.entries(finiteAutomaton.epsilon_transitions).forEach(([epsilonTransitionState, epsilonTransitionStates]) => {
                finiteAutomaton.epsilon_transitions[epsilonTransitionState] =
                    epsilonTransitionStates.map((epsilonTransitionState) => epsilonTransitionState.toString());
            });
        }
        function attachToStateRecord(transitionStateRecord, alphabetIndex, state) {
            if (!transitionStateRecord[finiteAutomaton.alphabets[alphabetIndex]]) {
                transitionStateRecord[finiteAutomaton.alphabets[alphabetIndex]] = [state];
            }
            else {
                transitionStateRecord[finiteAutomaton.alphabets[alphabetIndex]].push(state);
            }
        }
        Object.entries(finiteAutomaton.transitions).forEach(([transitionKey, transitionStates]) => {
            const transitionStateRecord = {};
            if (typeof transitionStates !== 'string' && Array.isArray(transitionStates)) {
                transitionStates.forEach((transitionState, transitionStateIndex) => {
                    if (transitionState !== null || transitionState !== undefined) {
                        if (Array.isArray(transitionState)) {
                            transitionState.forEach((state) => {
                                attachToStateRecord(transitionStateRecord, transitionStateIndex, appendedString + state.toString());
                            });
                        }
                        else {
                            attachToStateRecord(transitionStateRecord, transitionStateIndex, appendedString + transitionState.toString());
                        }
                    }
                });
                finiteAutomaton.transitions[appendedString + transitionKey] = transitionStateRecord;
                if (appendedString) {
                    delete finiteAutomaton.transitions[transitionKey];
                }
            }
            else if (transitionStates === 'loop') {
                finiteAutomaton.alphabets.forEach((_, alphabetIndex) => {
                    attachToStateRecord(transitionStateRecord, alphabetIndex, appendedString + transitionKey.toString());
                });
                finiteAutomaton.transitions[appendedString + transitionKey] = transitionStateRecord;
                if (appendedString) {
                    delete finiteAutomaton.transitions[transitionKey];
                }
            }
        });
        this.#validate(finiteAutomaton.label ?? '', this.#generatePostNormalizationErrors(finiteAutomaton));
        this.automaton = finiteAutomaton;
    }
    getAutomatonId() {
        return this.#automatonId;
    }
    #validate(automatonLabel, automatonValidationErrors) {
        if (automatonValidationErrors.length !== 0) {
            console.log(`${colors_1.default.blue.bold(automatonLabel)} ${colors_1.default.red.bold(automatonValidationErrors.length.toString())} Errors`);
            automatonValidationErrors.forEach((automatonValidationError) => console.log(colors_1.default.red.bold(automatonValidationError)));
            console.log();
            throw new Error(`Error validating automaton`);
        }
    }
    #generatePreNormalizationErrors(automaton) {
        const automatonValidationErrors = [];
        const { testLogic } = this;
        if (!testLogic) {
            automatonValidationErrors.push('testLogic function is required in Automaton');
        }
        if (!automaton.label) {
            automatonValidationErrors.push('Automaton label is required');
        }
        if (!automaton.states) {
            automatonValidationErrors.push('Automaton states is required');
            automaton.states = [];
        }
        if (!automaton.alphabets) {
            automatonValidationErrors.push('Automaton alphabets is required');
            automaton.alphabets = [];
        }
        if (!Array.isArray(automaton.states)) {
            automatonValidationErrors.push('Automaton states must be an array');
        }
        if (automaton.states.length === 0) {
            automatonValidationErrors.push('Automaton states must be an array of length > 0');
        }
        if (this.#automatonType === 'deterministic') {
            automaton.states.forEach((state) => {
                if (!automaton.transitions[state]) {
                    automatonValidationErrors.push(`Automaton states must reference a state (${state}) that is present in transitions`);
                }
            });
            if (automaton.epsilon_transitions) {
                automatonValidationErrors.push(`Deterministic automaton can't contain epsilon transitions`);
            }
        }
        if (automaton.start_state === undefined || automaton.start_state === null) {
            automatonValidationErrors.push('Automaton start_state is required');
        }
        if (automaton.final_states === undefined || automaton.final_states === null) {
            automatonValidationErrors.push('Automaton final_states is required');
            automaton.final_states = [];
        }
        if (!Array.isArray(automaton.final_states)) {
            automatonValidationErrors.push('Automaton final_states must be an array');
        }
        if (automaton.final_states.length === 0) {
            automatonValidationErrors.push('Automaton final_states must be an array of length > 0');
        }
        return automatonValidationErrors;
    }
    #generatePostNormalizationErrors(automaton) {
        const automatonValidationErrors = [];
        const automatonStates = new Set(automaton.states);
        const automatonAlphabets = new Set(automaton.alphabets);
        automaton.final_states.forEach((state) => {
            if (!automatonStates.has(state)) {
                automatonValidationErrors.push(`Automaton final_states must reference a state (${state}) that is present in states`);
            }
        });
        if (automaton.epsilon_transitions) {
            Object.entries(automaton.epsilon_transitions).forEach(([transitionState, transitionStates]) => {
                if (!automatonStates.has(transitionState)) {
                    automatonValidationErrors.push(`Epsilon transition state ${transitionState} must reference a state that is present in states`);
                }
                transitionStates.forEach((transitionState) => {
                    if (!automatonStates.has(transitionState)) {
                        automatonValidationErrors.push(`Epsilon transition state ${transitionState} must reference a state that is present in states`);
                    }
                });
            });
        }
        Object.entries(automaton.transitions).forEach(([transitionKey, transitionStatesRecord]) => {
            if (!automatonStates.has(transitionKey)) {
                automatonValidationErrors.push(`Automaton transitions (${transitionKey}) must reference a state that is present in states`);
            }
            const isTransitionValuesARecord = transitionStatesRecord &&
                typeof transitionStatesRecord === 'object' &&
                Object.getPrototypeOf(transitionStatesRecord) === Object.prototype;
            if (this.#automatonType === 'deterministic') {
                if (typeof transitionStatesRecord !== 'string' && !isTransitionValuesARecord) {
                    automatonValidationErrors.push(`Automaton transitions value must either be string "loop" or a tuple`);
                }
                if (typeof transitionStatesRecord === 'string' && transitionStatesRecord !== 'loop') {
                    automatonValidationErrors.push(`Automaton transitions value when a string, can only be "loop"`);
                }
            }
            if (transitionStatesRecord && typeof transitionStatesRecord !== 'string') {
                Object.entries(transitionStatesRecord).forEach(([transitionStateSymbol, transitionStateResultantStates]) => {
                    if (!automatonAlphabets.has(transitionStateSymbol)) {
                        automatonValidationErrors.push(`Automaton transitions symbol (${transitionStateSymbol}), must reference a valid alphabet`);
                    }
                    transitionStateResultantStates.forEach((transitionStateResultantState) => {
                        if (!automatonStates.has(transitionStateResultantState)) {
                            automatonValidationErrors.push(`Automaton transitions value (${transitionStateResultantState}) when a tuple, must reference a valid state`);
                        }
                    });
                });
            }
        });
        return automatonValidationErrors;
    }
    generateGraphFromString(inputString) {
        let currentParents = [
            {
                name: `${this.automaton.start_state}`,
                state: this.automaton.start_state,
                string: '',
                depth: 0,
                symbol: null,
                children: [],
            },
        ];
        const finalNodes = [];
        let automatonTestResult = false;
        const finalStates = new Set(this.automaton.final_states);
        const graph = currentParents;
        for (let index = 0; index < inputString.length; index++) {
            const newChildren = [];
            const symbol = inputString[index];
            currentParents.forEach((currentParent) => {
                const transitionState = this.automaton.transitions[currentParent.state];
                if (transitionState) {
                    const transitionStates = transitionState[symbol];
                    if (Array.isArray(transitionStates)) {
                        transitionStates.forEach((transitionState) => {
                            const parentGraphNode = {
                                name: transitionState + `(${symbol})`,
                                state: transitionState,
                                string: inputString.slice(0, index + 1),
                                depth: index + 1,
                                symbol,
                                children: [],
                            };
                            currentParent.children.push(parentGraphNode);
                            newChildren.push(parentGraphNode);
                        });
                    }
                }
            });
            if (index === inputString.length - 1) {
                for (let index = 0; index < newChildren.length; index++) {
                    const newChild = newChildren[index];
                    if (finalStates.has(newChild.state)) {
                        automatonTestResult = true;
                        break;
                    }
                }
                finalNodes.push(...newChildren);
            }
            currentParents = newChildren;
        }
        return {
            automatonTestResult,
            finalNodes,
            graph: graph[0],
        };
    }
}
exports.FiniteAutomaton = FiniteAutomaton;
