"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NonDeterministicFiniteAutomaton = void 0;
const DeterministicFiniteAutomaton_1 = require("./DeterministicFiniteAutomaton");
const FiniteAutomaton_1 = require("./FiniteAutomaton");
class NonDeterministicFiniteAutomaton extends FiniteAutomaton_1.FiniteAutomaton {
    constructor(testLogic, automaton, automatonId) {
        super(testLogic, automaton, 'non-deterministic', automatonId);
        if (this.automaton.epsilon_transitions) {
            this.convertToRegularNfa();
        }
    }
    convertToRegularNfa() {
        const { alphabets, epsilon_transitions: epsilonTransitions, transitions } = this.automaton;
        alphabets.forEach((alphabet) => {
            this.automaton.states.forEach((state) => {
                if (epsilonTransitions[state]) {
                    const firstPhaseSet = this.epsilonClosureOfState(state);
                    const secondPhaseSet = new Set();
                    const thirdPhaseSet = new Set();
                    firstPhaseSet.forEach((state) => {
                        if (transitions[state]) {
                            transitions[state][alphabet]?.forEach((transitionRecordState) => {
                                secondPhaseSet.add(transitionRecordState);
                            });
                        }
                    });
                    secondPhaseSet.forEach((secondPhaseState) => {
                        this.epsilonClosureOfState(secondPhaseState).forEach((state) => {
                            thirdPhaseSet.add(state);
                        });
                    });
                    if (this.automaton.transitions[state]) {
                        this.automaton.transitions[state][alphabet] = Array.from(thirdPhaseSet);
                    }
                    else {
                        this.automaton.transitions[state] = {
                            [alphabet]: Array.from(thirdPhaseSet),
                        };
                    }
                }
            });
        });
    }
    epsilonClosureOfState(state) {
        const { epsilon_transitions: epsilonTransitions } = this.automaton;
        const stack = [];
        const allEpsilonStates = new Set(state);
        stack.push(state);
        while (stack.length !== 0) {
            const currentState = stack.pop();
            epsilonTransitions[currentState]?.forEach((epsilonTransitionState) => {
                if (!allEpsilonStates.has(epsilonTransitionState)) {
                    stack.push(epsilonTransitionState);
                    allEpsilonStates.add(epsilonTransitionState);
                }
            });
        }
        return Array.from(allEpsilonStates);
    }
    moveAndEpsilonClosureStateSet(states, symbol) {
        const transitionRecordExtractedStates = new Set();
        states.forEach((state) => {
            const statesForSymbol = this.automaton.transitions[state]?.[symbol];
            if (statesForSymbol) {
                statesForSymbol.forEach((stateForSymbol) => transitionRecordExtractedStates.add(stateForSymbol));
            }
        });
        const finalStates = new Set(transitionRecordExtractedStates);
        if (this.automaton.epsilon_transitions) {
            transitionRecordExtractedStates.forEach((transitionRecordExtractedState) => {
                const epsilonClosureOfStates = this.epsilonClosureOfState(transitionRecordExtractedState);
                epsilonClosureOfStates.forEach((epsilonClosureOfState) => {
                    finalStates.add(epsilonClosureOfState);
                });
            });
        }
        return Array.from(finalStates);
    }
    convertToDeterministicFiniteAutomaton(dfaOptions) {
        const separator = dfaOptions.separator ?? ',';
        const startState = this.automaton.start_state;
        const newStartStates = this.automaton.epsilon_transitions
            ? this.epsilonClosureOfState(startState)
            : [this.automaton.start_state];
        const newStartStateString = newStartStates.sort().join(separator);
        const newStates = new Set();
        const unmarkedStates = [newStartStateString];
        const newTransitionsRecord = {};
        const totalAlphabets = this.automaton.alphabets.length;
        const newFinalStates = new Set();
        newStates.add(newStartStateString);
        let hasDeadState = false;
        while (unmarkedStates.length !== 0) {
            const currentStatesString = unmarkedStates.shift();
            this.automaton.alphabets.forEach((symbol, symbolIndex) => {
                const newStateString = this.moveAndEpsilonClosureStateSet(currentStatesString.split(separator), symbol)
                    .sort()
                    .join(separator);
                if (!newStates.has(newStateString) && newStateString) {
                    newStates.add(newStateString);
                    unmarkedStates.push(newStateString);
                }
                if (!newStateString) {
                    hasDeadState = true;
                    newStates.add(`Ø`);
                }
                if (!newTransitionsRecord[currentStatesString]) {
                    newTransitionsRecord[currentStatesString] = new Array(totalAlphabets).fill(null);
                }
                if (newStateString)
                    newTransitionsRecord[currentStatesString][symbolIndex] = [newStateString];
                else {
                    newTransitionsRecord[currentStatesString][symbolIndex] = [`Ø`];
                }
            });
        }
        this.automaton.final_states.forEach((finalState) => {
            newStates.forEach((newState) => {
                if (newState.includes(finalState)) {
                    newFinalStates.add(newState);
                }
            });
        });
        if (hasDeadState) {
            newTransitionsRecord['Ø'] = {};
            this.automaton.alphabets.forEach((symbol) => {
                newTransitionsRecord['Ø'][symbol] = ['Ø'];
            });
        }
        return new DeterministicFiniteAutomaton_1.DeterministicFiniteAutomaton(this.testLogic, {
            alphabets: this.automaton.alphabets,
            final_states: Array.from(newFinalStates),
            label: dfaOptions?.label ?? this.automaton.label,
            start_state: newStartStateString,
            states: Array.from(newStates),
            transitions: newTransitionsRecord,
            epsilon_transitions: null,
            description: dfaOptions?.description ?? this.automaton.description,
        });
    }
}
exports.NonDeterministicFiniteAutomaton = NonDeterministicFiniteAutomaton;
